/* bdb-gs.pn.c */

/* This program is the node processor software for a parallel version
of the Shur complement Gauss-Seidel iterative linear solver algorithm for
block-diagonal-bordered form sparse matrices.  */

/* special considerations for this implementation: 

1) Matrices are in block-diagonal-bordered form, but do not have to be
   symmetric.

2) any number of processors can be specified up to an including the
total number of processors on the CM-5 

3) it is assumed that ordering and load balancing are performed a
priori

*/
/*---------------------------------------------------------*/
#include <stdio.h>
#include <sys/types.h>
#include <math.h>
#include <malloc.h>
#include <cm/cmmd.h>
#include "bdb_struct_par_gs.h"
#include "bdb_gs_MSG_NUM.h"
/*---------------------------------------------------------*/
#define YES        1
#define NO         0
#define EPSILON    0.0000001
/*---------------------------------------------------------*/
Block            *blocks;      /* blocks[]     */
Diag             *diag;        /* diag[]       */
BlockRow         *blockrow;    /* blockrow[]   */
LowerBorder      *lborder;     /* lborder[]    */
LBorderRow       *lborderrow;  /* lborderrow[] */
double           *x0;          /* x0[]         */
double           *x1;          /* x1[]         */
double           *x;           /* x[]          */
double           *b;           /* b[]          */
int               n;           /* dimension of the entire block-diagonal matrix */
int               ndiag;       /* dimension of block-diagonal partition */
int               nblocks;     /* # of blocks             */
int               nblockrow;   /* dimension of blockrow   */
int               nlborder;    /* dimension of lborder    */
int               nlborderrow; /* dimension of lborderrow */
/*---------------------------------------------------------*/
LastBlockColor   *lbcolors;    /* lbcolors[]   */
LastDiag         *ldiag;       /* ldiag[]      */
LastBlockRow     *lblockrow;   /* lblockrow[]  */
int               ln;          /* dimension of the last block */
int               nldiag;
int               ncolors;
int               nlblockrow;
int               lastblockstart;
int               lastblockend;
CommLB          **commlb;
int              *nrowcolor;
int              *nupdates;

int               nprocs;
int               procnum;
int               niter;
int               bytesreceived;
int               bytessent;

UpdateLastBlock **updatelb; /* updatelb[][] */
UpdateLastBlock *lbrbuf;   /* lbrbuf[] */

typedef struct DBLObj{unsigned int i1;
                      unsigned int i2;} DBL;
/*---------------------------------------------------------*/

/* function prototypes */
void InputData();

void GSSolveTiming_1();
void GSSolve_1();
void GS_1();
void GSConverge_1();

void GSSolveTiming();
void GSSolve();
void GSDiagonalBlocks();
void GSUpdateLastBlock();
void GSUpdateLastBlockCMAML();
void StoreHandlerULBCMAML();
void GSLastBlock();
void SendHandlerCMMD();
void ReceiveHandlerCMMD();
void GSLastBlockCMAML();
void StoreHandlerLBCMAML();
void GSConverge();
void GSConvergeCMAML();
void StoreHandlerConCMAML();
void Node_Exit();

main(argc,argv)
  int argc;
  char **argv;
{
  int i;

  CMMD_enable_host();
  enable_diag_write();
  procnum = CMMD_self_address();

    printf("InputData() \n");
    fflush(stdout);
    InputData();

if (nprocs == 1)
  {
    printf("GSSolveTiming_1 \n");
    fflush(stdout);
    GSSolveTiming_1();

    printf("GSSolve_1 - niter=%d \n",niter);
    fflush(stdout);
    GSSolve_1();
  } /* end if nprocs == 1 */
else
  {
    printf("GSSolveTiming \n");
    fflush(stdout);
    GSSolveTiming();

    printf("GSSolve - niter=%d \n",niter);
    fflush(stdout);
    GSSolve();
  } /* end if nprocs > 1 */

} /* end main() */


/*<<<<<<<<<<<<<<<<<<<<---------->>>>>>>>>>>>>>>>>>>>*/
void InputData()
{
  int i,j,k,l,m,iproc,icolor,loc,bufsize,bytesrec;
  CommLB *commlbtmp;

/* (0) set up variables at this processor */ 
  CMMD_receive_bc_from_host(&bbuf,(12*sizeof(int)));
  n              = bbuf.n;
  ln             = bbuf.ln;
  nprocs         = bbuf.nprocs;
  niter          = bbuf.niter;
  ncolors        = bbuf.ncolors;
  lastblockstart = bbuf.lastblockstart;
  lastblockend   = bbuf.lastblockend;
printf("ln=%d - nprocs=%d - procnum=%d  \n",ln,nprocs,procnum);
fflush(stdout);
printf("lastblockend=%d - lastblockstart=%d \n",lastblockend,lastblockstart);
fflush(stdout);

if (procnum < nprocs)
 {

/* (1) receive the block identifer array of structures */
  CMMD_receive(CMMD_host_node(),RECEIVE_NPBLOCKS,&nblocks,sizeof(int));

printf("receive the block identifer array of structures --- nblocks=%d \n",nblocks);
fflush(stdout);

  blocks = (Block *) calloc((nblocks+1),sizeof(Block));
  if (blocks == NULL)
   {
    printf("can't allocate memory for blocks[] \n");
    Node_Exit(1);
   } /* end if */
  bufsize = ((nblocks+1)*sizeof(Block));
  CMMD_receive(CMMD_host_node(),RECEIVE_PBLOCKS,blocks,bufsize);
/*
if (procnum == 1) {
  for (i=0;i<=nblocks;i++) {
    printf("%d \t%d %d %d %d \n",i,
               blocks[i].BlockID,
               blocks[i].BlockSz,
               blocks[i].ProcNum,
               blocks[i].DiagPtr);
    fflush(stdout); }}
*/

/* (2) receive the matrix diagonal array of structures */
  CMMD_receive(CMMD_host_node(),RECEIVE_NP,&ndiag,sizeof(int));

printf("receive the matrix diagonal array of structures --- n=%d \n",n);
fflush(stdout);

  diag = (Diag *) calloc((ndiag+1),sizeof(Diag));
  if (diag == NULL)
   {
    printf("can't allocate memory for diag[] \n");
    Node_Exit(1);
   } /* end if */
  bufsize = ((ndiag+1)*sizeof(Diag));
  CMMD_receive(CMMD_host_node(),RECEIVE_PDIAG,diag,bufsize);
/*
if (procnum == 1) {
  for (i=0;i<=ndiag;i++) {
    printf("%d \t%lf %d \n",i,
             diag[i].Value,
             diag[i].PtrAix);
    fflush(stdout); }}
*/
/* (3) receive the block row storage array of structures */
  CMMD_receive(CMMD_host_node(),RECEIVE_NPBLOCKROW,&nblockrow,sizeof(int));

printf("receive the block row storage array of structures --- nblockrow=%d \n",nblockrow);
fflush(stdout);

  blockrow = (BlockRow *) calloc((nblockrow+1),sizeof(BlockRow));
  if (blockrow == NULL)
   {
    printf("can't allocate memory for blockrow[] \n");
    Node_Exit(1);
   } /* end if */
  bufsize = ((nblockrow+1)*sizeof(BlockRow));
  CMMD_receive(CMMD_host_node(),RECEIVE_PBLOCKROW,blockrow,bufsize);

printf("\treceived the block row storage array of structures --- nblockrow=%d \n",nblockrow);
fflush(stdout);

/*
if (procnum == 1) {
  for (i=0;i<=nblockrow;i++) {
    printf("%d \t%lf %d %d \n",i,
                    blockrow[i].Value,
                    blockrow[i].i,
                    blockrow[i].j);
    fflush(stdout); }}
*/

/* (4) receive the lower border row storage array of structures */
  CMMD_receive(CMMD_host_node(),RECEIVE_NPLBORDER,&nlborder,sizeof(int));

printf("receive the lower border row storage array of structures --- nlborder=%d \n",nlborder);
fflush(stdout);

  lborder = (LowerBorder *) calloc((nlborder+1),sizeof(LowerBorder));
  if (lborder == NULL)
   {
    printf("can't allocate memory for lborder[] \n");
    Node_Exit(1);
   } /* end if */
  bufsize = ((nlborder+1)*sizeof(LowerBorder));
  CMMD_receive(CMMD_host_node(),RECEIVE_PLBORDER,lborder,bufsize);
/*
if (procnum == 1) {
  for (i=0;i<=nlborder;i++) {
    printf("%d \t%d %d %d %d \n",i,
                   lborder[i].i,
                   lborder[i].ulbproc,
                   lborder[i].ulbloc,
                   lborder[i].PtrLBix);
    fflush(stdout); }}
*/

/* (5) receive the lower border row storage array of structures */
  CMMD_receive(CMMD_host_node(),RECEIVE_NPLBORDERROW,&nlborderrow,sizeof(int));

printf("receive the lower border row storage array of structures --- nlborderrow=%d \n",nlborderrow);
fflush(stdout);

  lborderrow = (LBorderRow *) calloc((nlborderrow+1),sizeof(LBorderRow));
  if (lborderrow == NULL)
   {
    printf("can't allocate memory for lborderrow[] \n");
    Node_Exit(1);
   } /* end if */
  bufsize = ((nlborderrow+1)*sizeof(LBorderRow));
  CMMD_receive(CMMD_host_node(),RECEIVE_PLBORDERROW,lborderrow,bufsize);
/*
if (procnum == 1) {
  for (i=0;i<=nlborderrow;i++) {
    printf("%d \t%lf %d %d \n",i,
                    lborderrow[i].Value,
                    lborderrow[i].i,
                    lborderrow[i].j);
    fflush(stdout); }}
*/
/* (7) receive last diagonal matrix colors */
printf("(7) receive the color identifer array of structures --- ncolors=%d \n",ncolors);
fflush(stdout);

  lbcolors = (LastBlockColor *) calloc((ncolors+1),sizeof(LastBlockColor));
  if (lbcolors == NULL)
   {
    printf("can't allocate memory for lbcolors[] \n");
    Node_Exit(1);
   } /* end if */
  bufsize = ((ncolors+1)*sizeof(LastBlockColor));
  CMMD_receive(CMMD_host_node(),RECEIVE_PLBCOLORS,lbcolors,bufsize);

printf("(7) receive the color identifer array of structures --- ncolors=%d \n",ncolors);
fflush(stdout);

  for (i=0;i<=ncolors;i++) {
    printf("\t\t%d \t%d %d %d %d %d \n",i,
               lbcolors[i].Color,
               lbcolors[i].ProcNum,
               lbcolors[i].Starti,
               lbcolors[i].DiagPtr,
               lbcolors[i].RowPtr);
    fflush(stdout); }

/* (8) receive the matrix last block diagonal array of structures */
printf("(8) receive the matrix last block diagonal array of structures --- ncolors=%d \n",ncolors);
fflush(stdout);
  CMMD_receive(CMMD_host_node(),RECEIVE_NPLDIAG,&nldiag,sizeof(int));

printf("receive the matrix last block diagonal array of structures --- nldiag=%d \n",nldiag);
fflush(stdout);

  ldiag = (LastDiag *) calloc((nldiag+1),sizeof(LastDiag));
  if (ldiag == NULL)
   {
    printf("can't allocate memory for ldiag[] \n");
    Node_Exit(1);
   } /* end if */
  bufsize = ((nldiag+1)*sizeof(LastDiag));
  CMMD_receive(CMMD_host_node(),RECEIVE_PLDIAG,ldiag,bufsize);
/*
if (procnum == 1) {
  for (i=0;i<=nldiag;i++) {
    printf("%d \t%lf %d \n",i,
             ldiag[i].Value,
             ldiag[i].PtrAix);
    fflush(stdout); }}
*/

/* (9) receive the last block row storage array of structures */

printf("receive the last block row storage array of structures \n");
fflush(stdout);

  CMMD_receive(CMMD_host_node(),RECEIVE_NPLBLOCKROW,&nlblockrow,sizeof(int));

printf("receive the last block row storage array of structures --- nlblockrow=%d \n",nlblockrow);
fflush(stdout);

  lblockrow = (LastBlockRow *) calloc((nlblockrow+1),sizeof(LastBlockRow));
  if (lblockrow == NULL)
   {
    printf("can't allocate memory for lblockrow[] \n");
    Node_Exit(1);
   } /* end if */

printf("receive the last block row storage array of structures (2) --- nlblockrow=%d \n",nlblockrow);
fflush(stdout);

  bufsize = ((nlblockrow+1)*sizeof(LastBlockRow));
  CMMD_receive(CMMD_host_node(),RECEIVE_PLBLOCKROW,lblockrow,bufsize);
/*
if (procnum == 1) {
  for (i=0;i<=nlblockrow;i++) {
    printf("%d \t%lf %d %d \n",i,
                    lblockrow[i].Value,
                    lblockrow[i].i,
                    lblockrow[i].j);
    fflush(stdout); }}
*/

/* (10) receive the sizes of messages to update the last block for each processor */

printf("receive the sizes of messages to update the last block for each processor --- nprocs=%d \n",nprocs);
fflush(stdout);

  ulbbufsize = (int **) calloc(nprocs,sizeof(int *));
  if (ulbbufsize == NULL)
   {
    printf("can't allocate memory for ulbbufsize[][] \n");
    Node_Exit(1);
   } /* end if */
  ulbbufsizetmp = (int *) calloc((nprocs*nprocs),sizeof(int));
  if (ulbbufsizetmp == NULL)
   {
    printf("can't allocate memory for ulbbufsizetmp[] \n");
    Node_Exit(1);
   } /* end if */
  for (i=0;i<nprocs;i++) ulbbufsize[i] = ulbbufsizetmp + (i * nprocs);

  bufsize = ((nprocs*nprocs)*sizeof(int));
  CMMD_receive(CMMD_host_node(),RECEIVE_PULBBUF,ulbbufsizetmp,bufsize);

if (procnum == 1) {
  for (i=0;i<nprocs;i++) {
   for (j=0;j<nprocs;j++) {
    printf("(%d,%d, %d) ",i,j,ulbbufsize[i][j]); }
   printf("\n");
    fflush(stdout); }}

/* (12) receive the b vector */
  b = (double *) calloc((n+ln+1),sizeof(double));
  if (b == NULL)
   {
    printf("can't allocate memory for b[] \n");
    Node_Exit(1);
   } /* end if */
  i = (blocks[0].DiagPtr + 1);
  bufsize = ((ndiag+1)*sizeof(double));

printf("receive the b vector - start location - i=%d - ndiag=%d - bufsize=%d - (n+ln)=%d \n",
               i,ndiag,bufsize,(n+ln));
fflush(stdout);

  CMMD_receive(CMMD_host_node(),RECEIVE_PB,&(b[i]),bufsize);

/* (13) receive the lb vector */
printf("(III) receive a portion of the lb vector \n");
fflush(stdout);
  bufsize = (ln*sizeof(double));
  CMMD_receive(CMMD_host_node(),RECEIVE_LB,&(b[lastblockstart]),bufsize);
/*
for (i=lastblockstart;i<=lastblockend;i++) {
  printf("\t\ti=%d - b[%d]=%lf \n",i,i,b[i]);
  fflush(stdout); }
*/
/* (16) receive the data structures for comm when solving for nodes in the last block */

printf("receive the data structures for comm when solving for nodes in the last block --- ncolors=%d - nprocs=%d \n",ncolors,nprocs);
fflush(stdout);

  commlbtmp = (CommLB *) calloc(((nprocs+1)*(ncolors+1)),sizeof(CommLB));
  if (commlbtmp == NULL)
   {
    printf("can't allocate memory for commlbtmp[] \n");
    exit(1);
   } /* end if */
  commlb = (CommLB **) calloc((ncolors+1),sizeof(CommLB *));
  if (commlb == NULL)
   {
    printf("can't allocate memory for commlb[][] \n");
    exit(1);
   } /* end if */
  for (icolor=0;icolor<=ncolors;icolor++) commlb[icolor] = commlbtmp + icolor*(nprocs+1);
  bufsize = ((nprocs+1)*(ncolors+1)*sizeof(CommLB));
  CMMD_receive(CMMD_host_node(),RECEIVE_PULBBUF,commlbtmp,bufsize);
/*
if (procnum == 1) {
  for (i=0;i<=ncolors;i++) {
   printf("%d \t",i);
   for (j=0;j<=nprocs;j++) {
     printf("\tcommlb[%d][%d].starti=%d - commlb[%d][%d].bufsize=%d \n",
              i,j,commlb[i][j].starti,i,j,commlb[i][j].bufsize); }
   printf("\n");
   fflush(stdout); } }
*/
  nrowcolor = (int *) calloc((ncolors+1),sizeof(int));
  if (nrowcolor == NULL)
   {
    printf("can't allocate memory for nrowcolor[] \n");
    Node_Exit(1);
   } /* end if */
  for (icolor=0;icolor<=ncolors;icolor++)
    {
     nrowcolor[icolor] = 0;
     for (iproc=1;iproc<=nprocs;iproc++)
       {
        nrowcolor[icolor] += commlb[icolor][iproc].bufsize;
       } /* end for iproc */
     printf("\ticolor=%d - nrowcolor[%d]=%d \n",icolor,icolor,nrowcolor[icolor]);
     fflush(stdout);
    } /* end for icolor */
  nupdates = (int *) calloc((nprocs+1),sizeof(int));
  if (nupdates == NULL)
   {
    printf("can't allocate memory for nupdates[] \n");
    Node_Exit(1);
   } /* end if */
  bufsize = (nprocs+1) * sizeof(int);
  CMMD_receive(CMMD_host_node(),RECEIVE_NUPDATE,nupdates,bufsize);

if (procnum == 1) {
  for (i=0;i<=nprocs;i++) {
     printf("iproc=%d - nupdates[%d]=%d \n",i,i,nupdates[i]);
     fflush(stdout); } }


/* (16) allocate space for the remaining vectors - sparse matrix software */
printf("allocate space for the remaining vectors - sparse matrix software \n");
fflush(stdout);

  x = (double *) calloc((n+ln+1),sizeof(double));
  if (x == NULL)
   {
    printf("can't allocate memory for x[] \n");
    Node_Exit(1);
   } /* end if */

  x0 = (double *) calloc((n+ln+1),sizeof(double));
  if (x0 == NULL)
   {
    printf("can't allocate memory for x0[] \n");
    Node_Exit(1);
   } /* end if */

  x1 = (double *) calloc((n+ln+1),sizeof(double));
  if (x1 == NULL)
   {
    printf("can't allocate memory for x1[] \n");
    Node_Exit(1);
   } /* end if */

  for (i=1;i<=(n+ln);i++) x[i] = x0[i] = x1[i] = 0.0;

  updatelb = (UpdateLastBlock **) calloc(nprocs,sizeof(UpdateLastBlock *));
  if (updatelb == NULL)
   {
    printf("can't allocate memory for updatelb[][] \n");
    Node_Exit(1);
   } /* end if */
  for (iproc=0;iproc<nprocs;iproc++)
    {
/*
printf("iproc=%d - ulbbufsize[%d][%d]=%d \n",iproc,procnum,iproc,ulbbufsize[procnum][iproc]);
fflush(stdout);
*/
      lbrbuf = (UpdateLastBlock *) calloc((ulbbufsize[procnum][iproc]+1),sizeof(UpdateLastBlock));
      if (lbrbuf == NULL)
       {
        printf("can't allocate memory for lbrbuf[] \n");
        Node_Exit(1);
       } /* end if */
     updatelb[iproc] = lbrbuf;
    } /* end for iproc */

  lbrbuf = (UpdateLastBlock *) calloc((ln+1),sizeof(UpdateLastBlock));
  if (lbrbuf == NULL)
    {
     printf("can't allocate memory for lbrbuf[] \n");
     Node_Exit(1);
    } /* end if */

/* preload the row values into updatelb[iproc][loc].i */

printf("preload the row values into updatelb[iproc][loc].i --- nlborder=%d \n",nlborder);
fflush(stdout);

  for (k=1; k<=nlborder; k++)
    {
      iproc = lborder[k].ulbproc;
      loc   = lborder[k].ulbloc;
      updatelb[iproc][loc].i = lborder[k].i;
/*
printf("k=%d - iproc=%d - loc=%d - updatelb[%d][%d].i=%d \n",
           k,iproc,loc,iproc,loc,updatelb[iproc][loc].i);
fflush(stdout);
*/
    } /* end for k */

 } /* end if this is an active processor */
} /* end InputData() */


/*<<<<<<<<<<<<<<<<<<<<---------->>>>>>>>>>>>>>>>>>>>*/ 
/* collect timing data on the sequential bdb-GS algorithm */
void GSSolveTiming_1()
{
  double res;

  CMMD_sync_with_host();
  if (procnum == 0) 
    {
printf("Gauss-Seidel - single processor \n");
fflush(stdout);
     CMMD_node_timer_clear(0);
     CMMD_node_timer_start(0);

     GS_1();

     CMMD_node_timer_stop(0);
     node_tim.pn_id   = CMMD_self_address();
     node_tim.elapsed = CMMD_node_timer_elapsed(0);
     node_tim.busy    = CMMD_node_timer_busy(0);
     node_tim.idle    = CMMD_node_timer_idle(0);
     CMMD_send(CMMD_host_node(),SEND_TIME_GSS1,&node_tim,sizeof(node_timer));
   } /* end if an active processor */

  CMMD_sync_with_host();
  if (procnum == 0) 
    {
printf("Gauss-Seidel - check convergence - 1 processor - procnum=%d\n",procnum);
fflush(stdout);
     CMMD_node_timer_clear(0);
     CMMD_node_timer_start(0);

     GSConverge_1(&res);

     CMMD_node_timer_stop(0);
     node_tim.pn_id   = CMMD_self_address();
     node_tim.elapsed = CMMD_node_timer_elapsed(0);
     node_tim.busy    = CMMD_node_timer_busy(0);
     node_tim.idle    = CMMD_node_timer_idle(0);
     CMMD_send(CMMD_host_node(),SEND_TIME_GSC1,&node_tim,sizeof(node_timer));

printf("\tthe residual - res=%lf \n",res);
fflush(stdout);

   } /* end if an active processor */

} /* end GSSolveTiming_1() */


/*<<<<<<<<<<<<<<<<<<<<---------->>>>>>>>>>>>>>>>>>>>*/ 
/* collect timing data on one cycle of the bdb-GS algorithm */
void GSSolve_1()
{
  int i,iter;
  double res,xmin,xmax;

  CMMD_sync_with_host();
  if (procnum == 0) 
    {
printf("Gauss-Seidel - solve with %d iterations and convergence check \n",niter);
fflush(stdout);
     CMMD_node_timer_clear(0);
     CMMD_node_timer_start(0);

     for (iter=0;iter<niter;iter++) GS_1();
     GSConverge_1(&res);
 
     CMMD_node_timer_stop(0);
     node_tim.pn_id   = CMMD_self_address();
     node_tim.elapsed = CMMD_node_timer_elapsed(0);
     node_tim.busy    = CMMD_node_timer_busy(0);
     node_tim.idle    = CMMD_node_timer_idle(0);
     CMMD_send(CMMD_host_node(),SEND_TIME_GST1,&node_tim,sizeof(node_timer));

printf("\tthe residual - res=%1.12lf \n",res);
fflush(stdout);

printf("the residual -------- \n");
fflush(stdout);

     for (i=1;i<=(n+ln);i++) x[i] = x0[i] = x1[i] = 0.0;

     for (iter=0;iter<32;iter++)
       {
        GS_1();
        GSConverge_1(&res);
printf("\t%d \t%d \t%1.12lf ",iter,(iter+1),res);
fflush(stdout);
        xmin = (double) (n+ln);
        xmax = 0.0;
        for (i=1;i<=(n+ln);i++)
          {
           if (x[i] > xmax) xmax = x[i];
           if (x[i] < xmin) xmin = x[i];
          } /* end for i */

printf("\txmin=%1.12lf \txmax=%1.12lf \n",xmin,xmax);
fflush(stdout);

       } /* end for iter*/

     for (i=1;i<=(n+ln);i++)
       {
printf("\tx[%d]=%1.12lf \tb[%d]=%1.12lf\n",i,x[i],i,b[i]);
fflush(stdout);
       } /* end for i */

   } /* end if an active processor */

  CMMD_sync_with_host();

} /* end GSSolve_1() */


/*<<<<<<<<<<<<<<<<<<<<---------->>>>>>>>>>>>>>>>>>>>*/ 
void GS_1()
{
  int i,j,k,l;
/*
printf("GS_1() --- blocks[0].DiagPtr=%d - blocks[1].DiagPtr=%d\n",
       blocks[0].DiagPtr,blocks[1].DiagPtr);
fflush(stdout);
*/
  for (i=(blocks[0].DiagPtr+1); i<=blocks[1].DiagPtr; i++)
   {
/*
printf("GS_1() --- i=%d - b[%d]=%lf \n",i,i,b[i]);
fflush(stdout);
*/
        x1[i] = b[i];
/* loop through all non-zero values in row i - diagonal block */
     for (j=(diag[i-1].PtrAix + 1);j<=diag[i].PtrAix;j++)
      {
/*
printf("\ti=%d - j=%d - x1[%d]=%lf - ",i,j,i,x1[i]);
printf("blockrow[%d].Value=%lf - ",j,blockrow[j].Value);
printf("blockrow[%d].j=%d -- ",j,blockrow[j].j);
printf("x[%d]=%lf \n",blockrow[j].j,x[blockrow[j].j]);
fflush(stdout);
*/
       x1[i] -= blockrow[j].Value * x[blockrow[j].j];
      } /* end for j */
    x[i] = x1[i] / diag[i].Value;
   } /* end for i - loop along diagonal in a block */
/*
printf("---------- exit GS_1() \n");
fflush(stdout);
*/
} /* end GS_1() */


/*<<<<<<<<<<<<<<<<<<<<---------->>>>>>>>>>>>>>>>>>>>*/ 
void GSConverge_1(res)
  double *res;
{
  int i;

/* calculate the total residual */
  *res = 0.0;
/* loop along the diagonal in a block */
  for (i=(blocks[0].DiagPtr+1); i<=blocks[1].DiagPtr; i++)
   {
    *res += fabs(x0[i]-x[i]);
    x0[i] = x[i];  /* set new x0[i] */
   } /* end for i - loop along diagonal in a block */
} /* end GSConverge_1() */


/*<<<<<<<<<<<<<<<<<<<<---------->>>>>>>>>>>>>>>>>>>>*/ 
/* collect timing data on each of four parts of the bdb-GS algorithm */
void GSSolveTiming()
{
  int icolor;
  double res;

  CMMD_sync_with_host();
  if (procnum < nprocs) 
    {
printf("Gauss-Seidel - diagonal blocks \n");
fflush(stdout);
     CMMD_node_timer_clear(0);
     CMMD_node_timer_start(0);

     GSDiagonalBlocks();

     CMMD_node_timer_stop(0);
     node_tim.pn_id   = CMMD_self_address();
     node_tim.elapsed = CMMD_node_timer_elapsed(0);
     node_tim.busy    = CMMD_node_timer_busy(0);
     node_tim.idle    = CMMD_node_timer_idle(0);
     CMMD_send(CMMD_host_node(),SEND_TIME_GSDB,&node_tim,sizeof(node_timer));
   } /* end if an active processor */

  CMMD_sync_with_host();
  if (procnum < nprocs) 
    {
printf("Gauss-Seidel - update last block \n");
fflush(stdout);
     CMMD_node_timer_clear(0);
     CMMD_node_timer_start(0);

     GSUpdateLastBlock();

     CMMD_node_timer_stop(0);
     node_tim.pn_id   = CMMD_self_address();
     node_tim.elapsed = CMMD_node_timer_elapsed(0);
     node_tim.busy    = CMMD_node_timer_busy(0);
     node_tim.idle    = CMMD_node_timer_idle(0);
     CMMD_send(CMMD_host_node(),SEND_TIME_GSULB,&node_tim,sizeof(node_timer));
   } /* end if an active processor */
 
  CMMD_sync_with_host();
  if (procnum < nprocs) 
    {
printf("Gauss-Seidel - last block \n");
fflush(stdout);
     CMMD_node_timer_clear(0);
     CMMD_node_timer_start(0);

     GSLastBlock();

     CMMD_node_timer_stop(0);
     node_tim.pn_id   = CMMD_self_address();
     node_tim.elapsed = CMMD_node_timer_elapsed(0);
     node_tim.busy    = CMMD_node_timer_busy(0);
     node_tim.idle    = CMMD_node_timer_idle(0);
     CMMD_send(CMMD_host_node(),SEND_TIME_GSLB,&node_tim,sizeof(node_timer));
   } /* end if an active processor */

  CMMD_sync_with_host();
  if (procnum < nprocs) 
    {
printf("Gauss-Seidel - check convergence \n");
fflush(stdout);
     CMMD_node_timer_clear(0);
     CMMD_node_timer_start(0);

     GSConverge(&res);

     CMMD_node_timer_stop(0);
     node_tim.pn_id   = CMMD_self_address();
     node_tim.elapsed = CMMD_node_timer_elapsed(0);
     node_tim.busy    = CMMD_node_timer_busy(0);
     node_tim.idle    = CMMD_node_timer_idle(0);
     CMMD_send(CMMD_host_node(),SEND_TIME_GSCC,&node_tim,sizeof(node_timer));

printf("\tthe residual - res=%lf \n",res);
fflush(stdout);

    } /* end if an active processor */

  CMMD_sync_with_host();
  CMAML_disable_interrupts();
  if (procnum < nprocs) 
    {
printf("Gauss-Seidel - update last block - active message version \n");
fflush(stdout);
     CMMD_node_timer_clear(0);
     CMMD_node_timer_start(0);

     GSUpdateLastBlockCMAML();

     CMMD_node_timer_stop(0);
     node_tim.pn_id   = CMMD_self_address();
     node_tim.elapsed = CMMD_node_timer_elapsed(0);
     node_tim.busy    = CMMD_node_timer_busy(0);
     node_tim.idle    = CMMD_node_timer_idle(0);
     CMMD_send(CMMD_host_node(),SEND_TIME_GSULBAML,&node_tim,sizeof(node_timer));
   } /* end if an active processor */
 
  CMMD_sync_with_host();
  if (procnum < nprocs) 
    {
printf("Gauss-Seidel - last block - active message version \n");
fflush(stdout);
     CMMD_node_timer_clear(0);
     CMMD_node_timer_start(0);

     GSLastBlockCMAML();

     CMMD_node_timer_stop(0);
     node_tim.pn_id   = CMMD_self_address();
     node_tim.elapsed = CMMD_node_timer_elapsed(0);
     node_tim.busy    = CMMD_node_timer_busy(0);
     node_tim.idle    = CMMD_node_timer_idle(0);
     CMMD_send(CMMD_host_node(),SEND_TIME_GSLBAML,&node_tim,sizeof(node_timer));
   } /* end if an active processor */
  else
    {
     for (icolor=1;icolor<=ncolors;icolor++)
       CMMD_sync_with_nodes();
    } /* end else */

  CMMD_sync_with_host();
  if (procnum < nprocs) 
    {
printf("Gauss-Seidel - check convergence - active message version \n");
fflush(stdout);
     CMMD_node_timer_clear(0);
     CMMD_node_timer_start(0);

     GSConvergeCMAML(&res);

     CMMD_node_timer_stop(0);
     node_tim.pn_id   = CMMD_self_address();
     node_tim.elapsed = CMMD_node_timer_elapsed(0);
     node_tim.busy    = CMMD_node_timer_busy(0);
     node_tim.idle    = CMMD_node_timer_idle(0);
     CMMD_send(CMMD_host_node(),SEND_TIME_GSCCAML,&node_tim,sizeof(node_timer));

printf("\tthe residual - res=%lf \n",res);
fflush(stdout);

    } /* end if an active processor */
   CMAML_enable_interrupts();
} /* end GSSolveTiming() */


/*<<<<<<<<<<<<<<<<<<<<---------->>>>>>>>>>>>>>>>>>>>*/ 
/* collect timing data on one cycle of the bdb-GS algorithm */
void GSSolve()
{
  int iter,icolor;
  double res;

  CMMD_sync_with_host();
  if (procnum < nprocs) 
    {
printf("Gauss-Seidel - solve with %d iterations and one convergence check \n",niter);
fflush(stdout);
     CMMD_node_timer_clear(0);
     CMMD_node_timer_start(0);

     for (iter=0;iter<niter;iter++)
       {
        GSDiagonalBlocks();
        GSUpdateLastBlock();
        GSLastBlock();
       } /* end for iter */
     GSConverge(&res);

     CMMD_node_timer_stop(0);
     node_tim.pn_id   = CMMD_self_address();
     node_tim.elapsed = CMMD_node_timer_elapsed(0);
     node_tim.busy    = CMMD_node_timer_busy(0);
     node_tim.idle    = CMMD_node_timer_idle(0);
     CMMD_send(CMMD_host_node(),SEND_TIME_GST,&node_tim,sizeof(node_timer));

printf("\tthe residual - res=%lf \n",res);
fflush(stdout);

   } /* end if an active processor */

  CMMD_sync_with_host();
  CMAML_disable_interrupts();
  if (procnum < nprocs) 
    {
printf("Gauss-Seidel - solve with %d iterations and one convergence check - active message version\n",niter);
fflush(stdout);
     CMMD_node_timer_clear(0);
     CMMD_node_timer_start(0);

     for (iter=0;iter<niter;iter++)
       {
        GSDiagonalBlocks();
        CMMD_sync_with_nodes();
        GSUpdateLastBlockCMAML();
        CMMD_sync_with_nodes();
        GSLastBlockCMAML();
        CMMD_sync_with_nodes();
       } /* end for iter */
     GSConvergeCMAML(&res);

     CMMD_node_timer_stop(0);
     node_tim.pn_id   = CMMD_self_address();
     node_tim.elapsed = CMMD_node_timer_elapsed(0);
     node_tim.busy    = CMMD_node_timer_busy(0);
     node_tim.idle    = CMMD_node_timer_idle(0);
     CMMD_send(CMMD_host_node(),SEND_TIME_GSTAML,&node_tim,sizeof(node_timer));

printf("\tthe residual - res=%lf \n",res);
fflush(stdout);

   } /* end if an active processor */
  else
    {
     for (iter=0;iter<niter;iter++)
       {
        CMMD_sync_with_nodes();
        CMMD_sync_with_nodes();
        for (icolor=1;icolor<=ncolors;icolor++)
          CMMD_sync_with_nodes();
        CMMD_sync_with_nodes();
       } /* end for iter */
    } /* end else */
  CMAML_enable_interrupts();

} /* end GSSolve() */


/*<<<<<<<<<<<<<<<<<<<<---------->>>>>>>>>>>>>>>>>>>>*/ 
void GSDiagonalBlocks()
{
  int i,j,k;

/* loop over the (nblocks) independent diagonal blocks */
/*
printf("GSDiagonalBlocks() --- nblocks=%d \n",nblocks);
fflush(stdout);
*/
/* loop along the diagonal in all blocks */
    for (i=(blocks[0].DiagPtr+1); i<=blocks[nblocks].DiagPtr; i++)
        {
         k = i - blocks[0].DiagPtr;
         x1[i] = b[i];
/*
printf("\ti=%d - k=%d - x1[%d]=%lf - b[%d]=%lf \n",i,k,i,x1[i],i,b[i]);
fflush(stdout);
*/
/* loop through all non-zero values in row i - diagonal block */
         for (j=(diag[ (k-1) ].PtrAix + 1);j<=diag[k].PtrAix;j++)
	   {
/*
if (procnum==1) {
printf("\ti=%d - x1[%d]=%lf - ",i,i,x1[i]);
printf("blockrow[%d].Value=%lf - ",j,blockrow[j].Value);
printf("blockrow[%d].j=%d -- ",j,blockrow[j].j);
printf("x[%d]=%lf \n",blockrow[j].j,x[blockrow[j].j]);
fflush(stdout); }
*/
            x1[i] -= blockrow[j].Value * x[blockrow[j].j];
	   } /* end for j */
/*
if (procnum==1) {
printf("\ti=%d x[%d]=%lf - x1[%d]=%lf - diag[%d].Value=%lf \n",
                 i,i,x[i],i,x1[i],k,diag[k].Value);
fflush(stdout); }
*/
         x[i] = x1[i] / diag[k].Value;
        } /* end for i - loop along diagonal in a block */

} /* end GSDiagonalBlocks() */


/*<<<<<<<<<<<<<<<<<<<<---------->>>>>>>>>>>>>>>>>>>>*/ 
void GSUpdateLastBlock()
{
  int i,j,k,l,loc,bufsize;
  int iproc,jproc;
/*
printf("In GSUpdateLastBlock() nprocs=%d \n",nprocs);
fflush(stdout); 
*/
  for (iproc=0;iproc<nprocs;iproc++)
    {
/*
printf("iproc=%d - ulbbufsize[%d][%d]=%d  \n",iproc,procnum,iproc,ulbbufsize[procnum][iproc]);
fflush(stdout); 
*/
      for (i=0; i<=ulbbufsize[procnum][iproc]; i++) 
      {
       updatelb[iproc][i].Value = 0.0;
      } /* end for i */
    } /* end for iproc */

/* reduce the border - except for the last lower diagonal block */
/* some lower border rows may be all zeros - so only use
   rows with nonzero values */
      for (k=1;k<=nlborder;k++)
       {
        iproc = lborder[k].ulbproc;
        loc   = lborder[k].ulbloc;
/*
if (procnum==1) {
printf("k=%d - iproc=%d - loc=%d \n",k,iproc,loc);
fflush(stdout); }
*/
        for (j=(lborder[k-1].PtrLBix + 1);j<=lborder[k].PtrLBix;j++)
	  {
/*
if (procnum==1) {
printf("\tj=%d -  lborderrow[%d].Value=%lf - lborderrow[%d].j=%d - x[%d]=%lf \n",
                 j,j,lborderrow[j].Value,j,lborderrow[j].j,
                 lborderrow[j].j,x[lborderrow[j].j]);
fflush(stdout); }
*/
           updatelb[iproc][loc].Value += lborderrow[j].Value * x[lborderrow[j].j];
/*
if (procnum==1) {
printf("\t\t\tupdatelb[%d][%d].i=%d - updatelb[%d][%d].Value=%lf  \n",
         iproc,loc,updatelb[iproc][loc].i,iproc,loc,updatelb[iproc][loc].Value);
fflush(stdout); }
*/
	  } /* end for j */
       } /* end for k */

/* send the appropriate data to the other processors - each processor 
   will send data to all other processors when it is it's turn - 
   otherwise the processor will receive.  After receiving data, 
   the processor updates it's values of x1( ).  After sending data, 
   a processor updates it's own values  */

  for (i=lastblockstart;i<=lastblockend;i++) x1[i] = b[i];
/*
printf("send the appropriate data to the other processors \n");
fflush(stdout);
*/
  for (iproc=0;iproc<nprocs;iproc++)
   {
    if (iproc == procnum) /* send updatelb[jproc] and update own lb( ) */
     {
      for (jproc=0;jproc<nprocs;jproc++)
       {
        if (jproc != procnum)
         {
          bufsize = (ulbbufsize[iproc][jproc] * sizeof(UpdateLastBlock));
/*
printf("CMMD_send --- source=%d - destination=%d - bufsize=%d - updatelb[%d][0].i=%d - updatelb[%d][0].Value=%lf \n",
               iproc,jproc,bufsize,jproc,updatelb[jproc][0].i,jproc,updatelb[jproc][0].Value);
fflush(stdout);
*/
           CMMD_send(jproc,(SEND_UPDATELB+jproc),updatelb[jproc],bufsize);
	 } /* end if */
       } /* end for jproc */
/* update x1[ ] stored on this processor */
      for (k=0;k<ulbbufsize[iproc][iproc];k++)
       {
/*
printf("k=%d -  i=%d - iproc=%d - updatelb[%d][%d].i=%d - updatelb[%d][%d].Value=%lf - ulbbufsize[%d][%d]=%d \n",
           k,i,iproc,iproc,k,updatelb[iproc][k].i,iproc,k,updatelb[iproc][k].Value,
           iproc,iproc,ulbbufsize[iproc][iproc]);
fflush(stdout); 
*/
        x1[ updatelb[iproc][k].i ] -= updatelb[iproc][k].Value;
/*
printf("\tiproc=%d - k=%d - updatelb[%d][%d].i=%d - updatelb[%d][%d].Value=%lf - x1[%d]=%lf \n",
       iproc,k,iproc,k,updatelb[iproc][k].i,iproc,k,updatelb[iproc][k].Value,
          updatelb[iproc][k].i,x1[ updatelb[iproc][k].i ]);
fflush(stdout); 
*/
       } /* end for k */
     } /* end for iproc == procnum (send data) */
    else /* receive lbrbuf and update own lb( ) */
     {
      bufsize = (ulbbufsize[iproc][procnum] * sizeof(UpdateLastBlock));
/*
printf("\tCMMD_receive --- source=%d - procnum=%d - bufsize=%d -  \n",iproc,procnum,bufsize);
fflush(stdout);
*/
      CMMD_receive(iproc,(RECEIVE_UPDATELB+procnum),lbrbuf,bufsize);
/*
printf("\tCMMD_receive --- lbrbuf[0].i=%d  - lbrbuf[0].Value=%lf \n",
             lbrbuf[0].i,lbrbuf[0].Value);
fflush(stdout);
*/
      for (k=0;k<ulbbufsize[iproc][procnum];k++)
       {
/*
printf("k=%d -  i=%d - iproc=%d - lbrbuf[%d].i=%d  \n",k,i,iproc,k,lbrbuf[k].i);
fflush(stdout); 
*/
        x1[ lbrbuf[k].i ] -= lbrbuf[k].Value;
/*
printf("k=%d - lbrbuf[%d].i=%d - lbrbuf[%d].Value=%lf - x1[%d]=%lf - ulbbufsize[%d][%d]=%d \n",
       k,k,lbrbuf[k].i,k,lbrbuf[k].Value,lbrbuf[k].i,x1[ lbrbuf[k].i ],
       iproc,procnum,ulbbufsize[iproc][procnum]);
fflush(stdout); 
*/
       } /* end for k */
     } /* end else (receive data) */

   } /* end for iproc */

} /* end GSUpdateLastBlock() */


/*<<<<<<<<<<<<<<<<<<<<---------->>>>>>>>>>>>>>>>>>>>*/
void GSUpdateLastBlockCMAML()
{
  int i,j,k,l;
  int iproc,jproc,ipoll;
  double update;
  unsigned int iupd1,iupd2;
  DBL iupd;
/*
printf("In GSUpdateLastBlockCMAML() - nprocs=%d - nupdates[%d]=%d \n",nprocs,(procnum+1),nupdates[procnum+1]);
fflush(stdout);
*/
  ipoll = 0;
  for (i=lastblockstart;i<=lastblockend;i++) x1[i] = b[i];

/* reduce the border - except for the last lower diagonal block */
/* some lower border rows may be all zeros - so only use
   rows with nonzero values */
  for (k=1;k<=nlborder;k++)
    {
     update = 0.0;
     i = lborder[k].i;
     iproc = lborder[k].ulbproc;
     for (j=(lborder[k-1].PtrLBix + 1);j<=lborder[k].PtrLBix;j++)
       {
/*
printf("\tj=%d -  lborderrow[%d].Value=%lf - lborderrow[%d].j=%d - x[%d]=%lf \n",
                 j,j,lborderrow[j].Value,j,lborderrow[j].j,
                 lborderrow[j].j,x[lborderrow[j].j]);
fflush(stdout); 
*/
         update += lborderrow[j].Value * x[lborderrow[j].j];

       } /* end for j */
/*
if (procnum<nprocs) {
printf("\tiproc=%d -  update=%lf - i=%d - ipoll=%d \n",iproc,update,i,ipoll);
fflush(stdout); }
*/
     if (iproc != procnum)
       {
        iupd1 = * ((int *)  (&update));
        iupd2 = * (((int *) (&update) + 1));
        CMAML_rpc(iproc,StoreHandlerLBCMAML,i,iupd1,iupd2,&ipoll);
       } /* end if */
     else
       {
        x[i] -= update;
        ipoll++;        
       } /* end else */
     CMAML_poll();
    } /* end for k */
/* check to see if data has been received */
  CMAML_poll_while_lt(&ipoll,nupdates[procnum+1]);
/*
printf("Exit GSUpdateLastBlockCMAML() - ipoll=%d - nupdates[%d]=%d \n",ipoll,(procnum+1),nupdates[procnum+1]);
fflush(stdout);
*/
} /* end GSUpdateLastBlockCMAML() */

/*<<<<<<<<<<<<<<<<<<<<---------->>>>>>>>>>>>>>>>>>>>*/
void StoreHandlerULBCMAML(i,iupd1,iupd2,ipoll)
  int i;
  unsigned int iupd1,iupd2;
  int *ipoll;
{
  DBL iupd;

  iupd.i1 = iupd1;
  iupd.i2 = iupd2;
  x[i] -= * ((double *) (&iupd));
  (*ipoll)++;
} /* end StoreHandlerULBCMAML() */

/*<<<<<<<<<<<<<<<<<<<<---------->>>>>>>>>>>>>>>>>>>>*/ 
void GSLastBlock()
{
  int i,j,k,l,icolor;
  int bufsize,iproc,jproc,BcastStatus;
  CMMD_mcb BcastMCB;
/*
printf("In GSLastBlock() ncolors=%d - nprocs=%d \n",ncolors,nprocs);
fflush(stdout); 
*/
/* nodes of the same color can be updated simultaniously  */
/* the last block has been ordered such that data for a processor
within a color are consecutive diagonal elements ---
as a result, the data can be sent without moving data
to and from buffers */
  for (icolor=1;icolor<=ncolors;icolor++)
    {
     for (k=(lbcolors[icolor-1].DiagPtr+1);k<=lbcolors[icolor].DiagPtr;k++)
       {
        i = ldiag[k].i;
        for (j=(ldiag[k-1].PtrAix+1);j<=ldiag[k].PtrAix;j++)
	  {
/*
printf("icolor=%d - k=%d - i=%d - j=%d - x1[%d]=%lf - lblockrow[%d].Value=%lf - lblockrow[%d].j=%d - x[%d]=%lf \n",
            icolor,k,i,j,i,x1[i],j,lblockrow[j].Value,j,lblockrow[j].j,
            lblockrow[j].j,x[lblockrow[j].j]);
fflush(stdout); 
*/
           x1[i] -= lblockrow[j].Value * x[lblockrow[j].j];
	  } /* end for j */
        x[i] = x1[i] / ldiag[k].Value;
/*
printf("\ti=%d - x1[%d]=%lf - x[%d]=%lf - ldiag[%d].Value=%lf \n",
            i,i,x1[i],i,x[i],k,ldiag[k].Value);
fflush(stdout); 
*/
       } /* end for k */
/*
printf("Communications phase ---------------------------------------------------- procnum=%d \n",procnum);
fflush(stdout);
*/
/* broadcast the new x[] vector values */
     for (iproc=0;iproc<nprocs;iproc++)
      {
       if (iproc == procnum) /* broadcast the portion of the updated x[] vector */
        {
/*
printf("broadcast the x data in this color on this processor to the other processors \n");
fflush(stdout);
*/
         for (jproc=0;jproc<nprocs;jproc++)
	   {
            if (jproc != procnum)
	      {
/*
printf("send --- icolor=%d -  jproc=%d - commlb[%d][%d].starti=%d - commlb[%d][%d].bufsize=%d \n",
         icolor,jproc,icolor,(iproc+1),commlb[icolor][iproc+1].starti,
                      icolor,(iproc+1),commlb[icolor][iproc+1].bufsize);
fflush(stdout); 
*/
               i = (commlb[icolor][iproc+1].starti + lastblockstart - 1);
               bufsize = (commlb[icolor][iproc+1].bufsize * sizeof(double));
               CMMD_send_async(jproc,(SEND_BCASTLX+iproc),&(x[i]),bufsize,SendHandlerCMMD,NULL);
	      } /* end if */
	   } /* for jproc */
        } /* end for iproc == procnum (send data) */
       else /* receive lx( ) */
        {
/*
printf("receive --- icolor=%d - iproc=%d - commlb[%d][%d].starti=%d - commlb[%d][%d].bufsize=%d \n",
         icolor,iproc,icolor,(iproc+1),commlb[icolor][iproc+1].starti,
                      icolor,(iproc+1),commlb[icolor][iproc+1].bufsize);
fflush(stdout); 
*/
          i = (commlb[icolor][iproc+1].starti + lastblockstart - 1);
          bufsize = (commlb[icolor][iproc+1].bufsize * sizeof(double));
/*
printf("\tCMMD_receive --- x[%d]=%lf - bufsize=%d\n",i,x[i],bufsize);
fflush(stdout);
*/
          BcastMCB = CMMD_receive_async(iproc,(RECEIVE_BCASTLX+iproc),&(x[i]),bufsize,ReceiveHandlerCMMD,&BcastStatus);
/* check to see if data has been received */
       if (BcastStatus != (RECEIVE_LB+iproc))
	 {
/*
printf("Wait to receive x[i] from processor - %d ---  \n",iproc);
fflush(stdout);
*/
           CMMD_msg_wait(BcastMCB);
	 } /* end if */
/*
printf("\tCMMD_receive --- x[%d]=%lf - bufsize=%d\n",i,x[i],bufsize);
fflush(stdout);
*/
        } /* end else (receive data) */
      } /* end for iproc */

    } /* end for icolor */

} /* end GSLastBlock() */

/*<<<<<<<<<<<<<<<<<<<<---------->>>>>>>>>>>>>>>>>>>>*/ 
void SendHandlerCMMD(mcb,arg)
  CMMD_mcb *mcb;
  void *arg;
{
  CMMD_free_mcb(*mcb);
} /* end SendHandlerCMMD() */

/*<<<<<<<<<<<<<<<<<<<<---------->>>>>>>>>>>>>>>>>>>>*/ 
void ReceiveHandlerCMMD(mcb,Status)
  CMMD_mcb *mcb;
  int *Status;  
{
  *Status - CMMD_mcb_tag(*mcb);
  CMMD_free_mcb(*mcb);
} /* end ReceiveHandlerCMMD() */


/*<<<<<<<<<<<<<<<<<<<<---------->>>>>>>>>>>>>>>>>>>>*/
void GSLastBlockCMAML()
{
  int i,j,k,l,icolor;
  int iproc,jproc,ipoll;
  unsigned int ix1,ix2;
  DBL ix;
/*
printf("In GSLastBlockCMAML() ncolors=%d - nprocs=%d - \n",ncolors,nprocs);
fflush(stdout);
*/
/* nodes of the same color can be updated simultaniously  */
/* the last block has been ordered such that data for a processor
within a color are consecutive diagonal elements ---
as a result, the data can be sent without moving data
to and from buffers */
  for (icolor=1;icolor<=ncolors;icolor++)
    {
     ipoll = 0;
     for (k=(lbcolors[icolor-1].DiagPtr+1);k<=lbcolors[icolor].DiagPtr;k++)
       {
        i = ldiag[k].i;
        for (j=(ldiag[k-1].PtrAix+1);j<=ldiag[k].PtrAix;j++)
         {
/*
printf("icolor=%d - k=%d - i=%d - j=%d - x1[%d]=%lf - lblockrow[%d].Value=%lf - lblockrow[%d].j=%d - x[%d]=%lf \n",
            icolor,k,i,j,i,x1[i],j,lblockrow[j].Value,j,lblockrow[j].j,
            lblockrow[j].j,x[lblockrow[j].j]);
fflush(stdout);
*/
           x1[i] -= lblockrow[j].Value * x[lblockrow[j].j];
          } /* end for j */
        x[i] = x1[i] / ldiag[k].Value;
        ix1 = * ((int *)  (&(x[i])));
        ix2 = * (((int *) (&(x[i]))) + 1);
        for (iproc=0;iproc<nprocs;iproc++)
          {
/*
printf("iproc=%d \n",iproc);
*/
           if (iproc != procnum)
             CMAML_rpc(iproc,StoreHandlerLBCMAML,i,ix1,ix2,&ipoll);
           else
             ipoll++;
          } /* end for iproc */
         CMAML_poll();
/*
printf("\ti=%d - x1[%d]=%lf - x[%d]=%lf - ldiag[%d].Value=%lf \n",
            i,i,x1[i],i,x[i],k,ldiag[k].Value);
fflush(stdout);
*/
       } /* end for k */
/* check to see if data has been received */
   CMAML_poll_while_lt(&ipoll,nrowcolor[icolor]);
/*
printf("End GSLastBlockCMAML() icolor=%d - ipoll=%d - nrowcolor[%d]=%d \n",
        icolor,ipoll,icolor,nrowcolor[icolor]);
fflush(stdout);
*/
     CMMD_sync_with_nodes();
    } /* end for icolor */
} /* end GSLastBlockCMAML() */

/*<<<<<<<<<<<<<<<<<<<<---------->>>>>>>>>>>>>>>>>>>>*/ 
void StoreHandlerLBCMAML(i,ix1,ix2,ipoll)
  int i;
  unsigned int ix1,ix2;
  int *ipoll;
{
  DBL ix;

  ix.i1 = ix1;
  ix.i2 = ix2;
  x[i]+= * ((double *) (&ix));
  (*ipoll)++;
} /* end StoreHandlerLBCMAML() */

/*<<<<<<<<<<<<<<<<<<<<---------->>>>>>>>>>>>>>>>>>>>*/ 
void GSConverge(gres)
  double *gres;
{
  int i,j,k,icolor;
  int nextproc,prevproc,AccumStatus;
  double res,restmp;
  CMMD_mcb AccumMCB;
/*
printf("In GSConverge() ncolors=%d - nprocs=%d \n",ncolors,nprocs);
fflush(stdout); 
*/
/* calculate the total residual */
  res = 0.0;
/* diagonal blocks on this processor */
/* loop along the diagonal in a block */
    for (i=(blocks[0].DiagPtr+1); i<=blocks[nblocks].DiagPtr; i++)
     {
/*
printf("i=%d - res=%lf - x0[%d]=%lf - x[%d]=%lf \n",i,res,i,x0[i],i,x[i]);
fflush(stdout);
*/
      res += fabs(x0[i]-x[i]);
      x0[i] = x[i];  /* set new x0[i] */
     } /* end for i - loop along diagonal in a block */

/* last diagonal block rows on this processor */
     for (icolor=1;icolor<=ncolors;icolor++)
       {
        j = (lbcolors[icolor-1].DiagPtr+1);
        k = lbcolors[icolor].DiagPtr;
        for (i=ldiag[j].i;i<=ldiag[k].i;i++)
          {
/*
printf("icolor=%d - i=%d - res=%lf - x0[%d]=%lf - x[%d]=%lf \n",icolor,i,res,i,x0[i],i,x[i]);
fflush(stdout);
*/
           res += fabs(x0[i]-x[i]);
           x0[i] = x[i];  /* set new x0[i] */
          } /* end for i */
       } /* end for icolor */

/* accumulate the total residual */
/*
printf("\tlocal contribution to the residual - res=%lf \n",res);
fflush(stdout);
*/
  nextproc = ((procnum+1) % nprocs);
  prevproc = ((procnum + nprocs - 1) % nprocs);
  *gres = res;
  for (i=0;i<(nprocs-1);i++)
    {
/*
printf("i=%d --- nextproc=%d - prevproc=%d \n",i,nextproc,prevproc);
*/
     CMMD_send_async(nextproc,SEND_ACCUM,&res,sizeof(double),SendHandlerCMMD,NULL);
     AccumMCB = CMMD_receive_async(prevproc,RECEIVE_ACCUM,&restmp,sizeof(double),ReceiveHandlerCMMD,&AccumStatus);
/* check to see if data has been received */
     if (AccumStatus != RECEIVE_ACCUM)
      {
/*
printf("Wait to receive x[i] from processor - %d ---  \n",prevproc);
fflush(stdout);
*/
        CMMD_msg_wait(AccumMCB);
      } /* end if */
     *gres += restmp;
     res = restmp;
/*
printf(" \tglobal residual - gres=%lf --- temp residual - restmp=%lf\n",*gres,restmp);
fflush(stdout);
*/
    } /* end for i */

} /* end GSConverge() */


/*<<<<<<<<<<<<<<<<<<<<---------->>>>>>>>>>>>>>>>>>>>*/ 
void GSConvergeCMAML(gres)
  double *gres;
{
  int i,j,k,l,icolor,iproc,ipoll;
  double res;
  unsigned int ires1,ires2;
  DBL ires;
/*
printf("In GSConvergeCMAML() ncolors=%d - nprocs=%d \n",ncolors,nprocs);
fflush(stdout); 
*/
/* calculate the total residual */
  res = 0.0;
  ipoll = 0;
/* diagonal blocks on this processor */
/* loop along the diagonal in a block */
    for (i=(blocks[0].DiagPtr+1); i<=blocks[nblocks].DiagPtr; i++)
     {
/*
printf("i=%d - res=%lf - x0[%d]=%lf - x[%d]=%lf \n",i,res,i,x0[i],i,x[i]);
fflush(stdout);
*/
      res += fabs(x0[i]-x[i]);
      x0[i] = x[i];  /* set new x0[k] */
     } /* end for i - loop along diagonal in a block */

/* last diagonal block rows on this processor */
     for (icolor=1;icolor<=ncolors;icolor++)
       {
        j = (lbcolors[icolor-1].DiagPtr+1);
        k = lbcolors[icolor].DiagPtr;
        for (i=ldiag[j].i;i<=ldiag[k].i;i++)
          {
/*
printf("icolor=%d - i=%d - res=%lf - x0[%d]=%lf - x[%d]=%lf \n",icolor,i,res,i,x0[i],i,x[i]);
fflush(stdout);
*/
           res += fabs(x0[i]-x[i]);
           x0[i] = x[i];  /* set new x0[i] */
          } /* end for i */
       } /* end for icolor */

/* accumulate the total residual */
/*
printf("\tlocal contribution to the residual - res=%lf \n",res);
fflush(stdout);
*/
  *gres = res;
  ires1 = * ((int *) (&res));
  ires2 = * (((int *) (&res)) + 1);
  for (iproc=0;iproc<nprocs;iproc++)
    {
/*
printf("iproc=%d \n",iproc);
*/
     if (iproc != procnum)
       CMAML_rpc(iproc,StoreHandlerConCMAML,gres,ires1,ires2,&ipoll);
    } /* end for i */
/* check to see if data has been received */
   CMAML_poll_while_lt(&ipoll,(nprocs-1));
/*
printf(" \tglobal residual - gres=%lf \n",*gres);
fflush(stdout);
*/
} /* end GSConvergeCMAML() */

/*<<<<<<<<<<<<<<<<<<<<---------->>>>>>>>>>>>>>>>>>>>*/ 
void StoreHandlerConCMAML(gres,ires1,ires2,ipoll)
  double *gres;
  unsigned int ires1,ires2;
  int *ipoll;
{
  DBL ires;

  ires.i1 = ires1;
  ires.i2 = ires2;
  *gres += * ((double *) (&ires));
  (*ipoll)++;
} /* end SendHandlerCMMD() */

/*<<<<<<<<<<<<<<<<<<<<---------->>>>>>>>>>>>>>>>>>>>*/
void Node_Exit(i)
  int i;
{
  int inf = 1073741824;
  double e1;

  e1 = b[inf];
  exit(i);

} /* Node_Exit() */
